--[[

This file is part of RampageAI.

RampageAI is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

RampageAI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RampageAI; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

--]]

---------------------------
-- Monster Table Support --
---------------------------

-- Convert the MonsterTypeDB and MonsterSpecDB tables
function ConvertMDB()
	-- Log it
	Log(0,"Optimizing monster database...")
	
	-- Verify we have a valid default
	if type(DefaultMonsterSettings) == "table" then
		-- Make sure the "ID" will not be counted as the end of the table
		DefaultMonsterSettings[1] = 0
	else
		-- Generate a blank table
		DefaultMonsterSettings = { }
	end

	-- Verify our default settings
	DefaultMonsterSettings = VerifyMonSettings(DefaultMonsterSettings,nil,"default")

	-- Function to convert from Monsters.lua table type into internal format
	local func = function(table)
		-- Make a copy of the monster db
		local mdb = { }
	
		-- Loop through each monster type
		local i,t
		for i,t in pairs(table) do
			-- Grab the ID
			local id = t[1]
	
			-- Verify the ID
			if type(id) == "number" then
				-- Add it to our new database
				--mdb[id] = VerifyMonSettings(t,false,tostring(id))
	
				-- Add it to our new database
				mdb[id] = t
	
				-- Make sure "verified" is false
				mdb[id].verified = false
			end
		end

		return mdb
	end

	MonsterTypeDB = func(MonsterTypeDB)
	MonsterSpecDB = func(MonsterSpecDB)
	
	-- Ensure this isn't repeated
	ConvertMDB = NoReturn
end

-- Verify monster settings
function VerifyMonSettings(t,defaultTable,id)
	-- Parameter names
	local argNames = { nil, "prio", "atk", "skLvl", "skDly", "skMax", "aSkLv", "aSkDl",
			"aSkMx", "defOnly", "kite", "ffa", "hlBm", "mammo", "aTer", "aBmb" }

	-- Check if we're working on the default
	local def
	if not defaultTable then
		def = { 0, 1, true, 5, 500, 1, 5, 1000, -1, false, false, false, false, 0, 0, 0 }
	else
		def = defaultTable
	end

	-- Check if they have a high enough level
	if DebugLevel >= 8 then
		-- Show them what the monster entry looked like before
		local j,k,str = 0,0,StringBuffer.New()
		StringBuffer.Append(str,string.format("Before: MDB[%s] = { ",id))
		for j,k in pairs(t) do
			if j ~= "verified" then
				StringBuffer.Append(str,string.format("%s=%q, ",
					tostring(j), tostring(k)))
			end
		end
		Log(8,"%s },",StringBuffer.Get(str))
	end

	-- Rename each "named" parameter to its corresponding type
	local i,val
	for i,val in pairs(argNames) do
		-- Check to see if the table has a parameter matching name
		if t[val] ~= nil then
			-- Set the corresponding argument
			t[i] = t[val]

			-- Remove the trace
			t[val] = nil
		end

		-- Make sure there is an option at all
		if t[i] == nil and def[i] ~= nil then
			-- Log it
			Log(8,"Missing option %q",tostring(i))

			-- Set the option
			t[i] = def[i]
		end
	end


	-- Loop through each value of the monster table
	for i,val in pairs(t) do
		-- Check if theres a default type for this option
		if def[i] ~= nil and type(val) ~= type(def[i]) then

			-- Log it
			Log(8,"Invalid option %q",
				tostring(i),id)

			-- Wrong type, so set it to our default
			t[i] = def[i]

		end
	end

	-- Check if they have a high enough level
	if DebugLevel >= 8 then
		-- Show them what the monster entry looks like afterwards
		local j,k,str = 0,0,StringBuffer.New()
		StringBuffer.Append(str,string.format("After: MDB[%s] = { ",id))
		for j,k in pairs(t) do
			if j ~= "verified" then
				StringBuffer.Append(str,string.format("%s=%q, ",
					tostring(j), tostring(k)))
			end
		end
		Log(8,"%s },",StringBuffer.Get(str))
	end

	-- This table was verified
	t.verified = true

	-- Return our verified table
	return t
end

-- Get the monster-specific table
function MDBTable(actor,useTypeOnly)
	-- Make sure the actor isn't nil
	if actor == nil then
		-- Return default
		return DefaultMonsterSettings
	end

	-- Get the actor type
	local type = ActorData[actor].Type

	-- Variable to hold the table
	local t = MonsterSpecDB[actor]

	-- Check if there is a specific monster table
	if t ~= nil and not useTypeOnly then
		-- Its a specific monster, grab its table

		-- Check if the table has been verified yet
		if not t.verified then
			-- Log it
			Log(2,"First access of specific monster settings for Monster ID #%d, verifying...",
				actor)

			-- Verify it (using its type for default settings)
			MonsterSpecDB[actor] = VerifyMonSettings(t,MDBTable(actor,true),tostring(actor))

			-- Regrab the table (in case it changes for some reason)
			t = MonsterSpecDB[actor]
		end

	elseif MonsterTypeDB[type] ~= nil then
		-- It is, grab the monster-type table
		t = MonsterTypeDB[type]

		-- Check if its been verified yet
		if not t.verified then
			-- Log it
			Log(2,"First access of monster-type settings for Monster Type ID #%d, verifying...",
				type)

			-- Verify it
			MonsterTypeDB[type] = VerifyMonSettings(t,DefaultMonsterSettings,tostring(type))

			-- Regrab the table (in case it changes for some reason)
			t = MonsterTypeDB[type]
		end
	else
		-- Not specific, and not type. Return the default
		return DefaultMonsterSettings
	end

	-- Return the table
	return t
end

-------------------------------
-- Monster Selection Support --
-------------------------------

-- Check if we're allowed to target the monster
function MonsterAllowed(actor,foralch)
	-- Check if either skills or melee are allowed against the monster
	if not foralch and (MSkillsAllowed(actor) or MAttackAllowed(actor)) then
		return true
	elseif foralch and MAlchSkAllowed(actor) then
		return true
	end
	
	return false
end

-- Check if we are allowed to melee the monster
function MAttackAllowed(actor)
	-- Get the monster's table
	local t = MDBTable(actor)

	-- Return the melee allowed value
	return t[3]
end

-- Check if the monster is free-for-all, and can be KSed at any time
--	** idea from HauNix of RagnaInfo forums **
function MIgnoreKS(actor)
	-- Get the monster's table
	local t = MDBTable(actor)

	-- Return the value
	return t[12]
end

-- Check if we're supposed to target this monster only if it attacks us
function MNoAggroAgainst(actor)
	-- Get the monster's table
	local t = MDBTable(actor)

	-- Return the defend only value
	return t[10]
end

-- Check if we're supposed to kite this monster rather than attack it
function MKite(actor)
	-- Get the monster's table
	local t = MDBTable(actor)

	-- Return the kite value
	return t[11]
end

-- Function to remove all the monsters that are supposed to be kited,
--	unless they're all kite monsters
function MKiteSieve(t)
	local max = Table.getn(t)

	-- Check if we have any
	if max == 0 then
		-- Return a new table
		return Table.new()
	end

	-- Setup the default table
	local tablekite = true
	local monsters = Table.new()

	local i,actor,monkite
	for i=1,max do
		-- Get the actor
		actor = t[i]

		-- Check if the monster should be kited
		monkite = MKite(actor)

		-- Check if our table is currently full of kitable monsters
		if tablekite and not monkite then
			-- We are no longer filling with kitable monsters
			tablekite = false

			-- Make a new table
			monsters = Table.new()
		end

		-- Check if we should add to the table
		if tablekite == monkite then
			-- Add to the table
			Table.insert(monsters,actor)
		end
	end

	-- Return the table of monsters
	return monsters
end

-- Check the priority level of a monster
function MAttackPriority(actor)
	-- Check the mosnter's type
	local t = MDBTable(actor)
	
	-- Return the priority level
	return t[2]
end

-- Function to select the highest priority monsters from a list
function MPrioritySieve(t)
	-- Get the highest value of the table
	local max = Table.getn(t)
	
	-- Check if we have none
	if max == 0 then
		return Table.new()
	end

	-- Setup the default table
	local priority = MAttackPriority(t[1])
	local monsters = Table.new()
	
	-- Loop through our table, sorting the highest priority monsters into our new table
	local i,actor,prio
	for i=1,max do
		-- Get the actor
		actor = t[i]
		
		-- Check the actor's priority
		prio = MAttackPriority(actor)
		
		-- Check if we should start a new priority table
		if prio > priority then
			-- Priority is higher, start a new table
			priority = prio
			monsters = Table.new()
		end
		
		-- Check if we should add the monster to the table
		if prio == priority then
			-- Add the monster
			Table.insert(monsters,actor)
		end
	end
	
	-- Return the table with the highest priority monsters
	return monsters
end

------------------------------
-- Homunculus Skill Support --
------------------------------

-- Check if we are allowed to use skills against the monster
function AI_MSkillsAllowed(actor)
	-- Check if we're pausing skills for the moment
	if PauseAttackSkill then
		-- They're paused
		return false
	end

	-- Get the monster's table
	local t = MDBTable(actor)

	-- Check if level is greater than 0 and we haven't casted too often
	--	against them
	if t[4] > 0 and MSkillMaxCasts(actor) > ActorData[actor].Count.HomuSkill then
		return true
	end

	-- If the value isn't greater than 0, disable skill usage
	return false
end

-- Check the monster-specific skill timeout
function MSkillTimeout(actor)
	-- Get the monster's table
	local t = MDBTable(actor)
	
	-- Return the timeout value
	return t[5]
end

-- Get the maximum level of an attack skill to use against a monster
function MSkillMaxLevel(actor)
	-- Get the monster's table
	local t = MDBTable(actor)

	-- Return the max skill level value
	return t[4]
end

-- Get the maximum number of times to cast a homunculus skill against the monster
function MSkillMaxCasts(actor)
	-- Get the monster's table
	local t = MDBTable(actor)

	-- Get the max number of casts
	local ret = t[6]

	-- Check if it is negative
	if ret < 0 then
		--[[
			Return 1 higher than our current number of casts
			against our target (so that its seemingly infinite)
		--]]

		ret = ActorData[actor].Count.HomuSkill + 1
	end

	-- Return the max number of casts
	return ret
end

------------------------------------
-- Auto-Alch Attack Skill Support --
------------------------------------

-- Check if we are allowed to cast against the monster
function MAlchSkAllowed(actor)
	-- Get the monster's table
	local t = MDBTable(actor)

	-- Check if skills are allowed
	if (t[7] > 0 and MAlchSkMaxCasts(actor) > ActorData[actor].Count.AlchSkill) or
		MAlchSkMammo(actor) > ActorData[actor].Count.AlchMammo or
		MAlchSkAcidTerror(actor) > ActorData[actor].Count.AlchAcidT or
		MAlchSkAcidBomb(actor) > ActorData[actor].Count.AlchAcidB then
		return true
	end

	-- They aren't
	return false
end

-- Get the timeout against this monster
function MAlchSkTimeout(actor)
	-- Get the monster's table
	local t = MDBTable(actor)

	-- Return the timeout
	return t[8]
end

-- Get the max level of bolt to cast
function MAlchSkMaxLevel(actor)
	-- Get the monster's table
	local t = MDBTable(actor)

	-- Return the level
	return t[7]
end

-- Get the max number of casts against the monster
function MAlchSkMaxCasts(actor)
	-- Get the monster's table
	local t = MDBTable(actor)

	-- Get the max number of casts
	local ret = t[9]

	-- Check if its negative
	if ret < 0 then
		--[[
			Negative means infinite, so add 1 to our
			current count of casts
		--]]

		ret = ActorData[actor].Count.AlchSkill + 1
	end

	return ret
end

-- Determine if the monster should be heal-bombed
function MAlchSkHealBomb(actor)
	-- Get the monster's table
	local t = MDBTable(actor)

	-- Return the heal-bomb value
	return t[13]
end

-- Determine if the monster should be mammonited
function MAlchSkMammo(actor)
	-- Get the monster's table
	local t = MDBTable(actor)

	-- Get the mammonite value
	local ret = t[14]

	-- Check if its negative
	if ret < 0 then
		-- Set it higher
		ret = ActorData[actor].Count.AlchMammo + 1
	end

	return ret
end

-- Determine if the monster should be acid-terrored
function MAlchSkAcidTerror(actor)
	-- Get the monster's table
	local t = MDBTable(actor)

	-- Get the mammonite value
	local ret = t[15]

	-- Check if its negative
	if ret < 0 then
		-- Set it higher
		ret = ActorData[actor].Count.AlchAcidT + 1
	end

	return ret
end

-- Determine if the monster should be acid-demonstrationed
function MAlchSkAcidBomb(actor)
	-- Get the monster's table
	local t = MDBTable(actor)

	-- Get the mammonite value
	local ret = t[16]

	-- Check if its negative
	if ret < 0 then
		-- Set it higher
		ret = ActorData[actor].Count.AlchAcidB + 1
	end

	return ret
end

----------------------------
-- Monster Ignore Support --
----------------------------

-- Remove all the ignored monsters from a Table
function IgnoredSieve(t)
	-- Reuse code from the table set
	return Table.sieveValues(t,IsIgnored)

--[[
	-- Loop through each monster in the table
	local i,max = 1,Table.getn(t)
	while i <= max do
		-- Check if the monster is ignored
		if IsIgnored(t[i]) then
			-- Monster is ignored, remove it
			Table.remove(t,i)

			-- Reduce the max
			max = Table.getn(t)
		else
			-- Monster isn't ignored, move to next
			i = i + 1
		end
	end

	return t
--]]
end

-- Check our list of ignored monsters
function CheckIgnored(ticks)
	-- Loop through all of our ignored monsters
	local actor,time
	for actor,time in pairs(IgnoredTargets) do
		-- Check if its ignored
		if IgnoredTargets[actor] > 0 then
			-- Subtract the time from the target
			IgnoredTargets[actor] = IgnoredTargets[actor] - ticks
	
			-- Log it
			Log(3,"Removed %d ticks from Actor #%d's ignore time, resulting in %d",
				ticks, actor, IgnoredTargets[actor])
		end
	end
end

-- Check if a monster is ignored
function IsIgnored(actor)
	-- Loop through our ignored monsters
	if type(IgnoredTargets[actor]) == "number" and
		IgnoredTargets[actor] > 0 then
		-- They still have time left
		return true
	end

	-- If its not in the ignored list, its not ignored
	return false
end

-- Add a monster to the ignore list
function AddIgnore(actor,time)
	-- Check if its already ignored
	if IsIgnored(actor) then
		-- Do nothing
		return
	end

	-- Check the time
	if type(time) ~= "number" then
		time = ActorIgnoreTime
	end
	
	-- Log it
	Log(0,"Ignoring actor #%d for %d milliseconds.",actor,time)

	-- Add the actor to the ignore list
	IgnoredTargets[actor] = time
end